"""Test - Library
"""

import asyncio
import functools
import os
import pathlib
import sys

import pytest


def dir_up(depth):
    """Easy level-up folder(s)."""
    return sys.path.append(os.path.join(pathlib.Path(__file__).parents[depth], "src"))


# Append to (sys.path)
dir_up(1)


import functools

# Test
from sqlalchemy import create_engine

import dbcontroller as dbc

sql = dbc.Controller(sql="sqlite:///example.db")

# Setup
engine = create_engine(sql.url, echo=True)


# Types
@sql.model(table_name="main_user")
class UserSQL:
    name: str
    notes: dbc.text
    meta: dbc.json
    disabled: bool = False

# Init Objects
dbc.load([UserSQL])
class ID:
    one = "MTo6YTU1ZTUzMmVhYjAyOGI0Mg=="
    two = "Mjo6M2VmOWFiYmI1ZGY1YjY0MQ=="


def test_create_database():
    sql.base.metadata.create_all(engine)
    assert True


@pytest.mark.asyncio
async def test_create():
    results = await UserSQL.objects.create({"name": "jane doll"})
    assert results.error == False and results.data.name == "jane doll"


@pytest.mark.asyncio
async def test_update():
    selector = ID.one
    form = {
        "name": "joe doe",
    }
    results = await UserSQL.objects.update(selector, form)
    assert results.error == False and results.data.name == "joe doe"


@pytest.mark.asyncio
async def test_detail():
    selector = ID.one
    results = await UserSQL.objects.detail(selector)
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_get_by():
    results = await UserSQL.objects.get_by(id=1)
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_find_one():
    query = UserSQL.objects.where("name", "contains", "joe")
    results = await UserSQL.objects.find_one(query)
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_delete():
    selector = ID.one
    results = await UserSQL.objects.delete(selector)
    assert results.error == False and results.count == 1


@pytest.mark.asyncio
async def test_all():
    await UserSQL.objects.create([{"name": "joe doe"}, {"name": "jane doll"}])
    results = await UserSQL.objects.all()
    assert (
        results.error == False
        and isinstance(results.data, list) == True
        and len(results.data) == 2
    )


@pytest.mark.asyncio
async def test_filter_by():
    query = {"name": "joe doe"}
    results = await UserSQL.objects.filter_by(search=query, page=1, limit=100, sort_by="-id")
    assert results.error == False and results.count == 1


@pytest.mark.asyncio
async def test_find():
    query = UserSQL.objects.where("name", "contains", "jane") | UserSQL.objects.where(
        "name", "contains", "joe"
    )
    results = await UserSQL.objects.find(query, page=1, limit=100, sort_by="-id")
    assert results.error == False and results.count == 2


@pytest.mark.asyncio
async def test_search():
    search_value = "j"
    columns = ["name", "notes"]
    results = await UserSQL.objects.search(
        columns=columns, value=search_value, page=1, limit=100, sort_by="-id"
    )
    assert results.error == False and results.count == 2


@pytest.mark.asyncio
async def test_reset():
    results = await UserSQL.objects.delete(None, all=True)
    assert results.error == False


async def demo_example():
    query = UserSQL.objects.where("name", "contains", "joe")
    results = await UserSQL.objects.find_one(query)
    assert results.name == "joe doe"


# asyncio.run(demo_example())



"""Test - Library
"""

import asyncio
import os
import pathlib
import sys

import pytest


def dir_up(depth):
    """Easy level-up folder(s)."""
    return sys.path.append(os.path.join(pathlib.Path(__file__).parents[depth], "src"))


# Append to (sys.path)
dir_up(1)

# Test
import dbcontroller as dbc

# Config
mongo = dbc.Controller(mongo="mongodb://localhost:27017/test_database")

# Types
@mongo.model(table_name="main_user")
class UserMongo:
    name: str
    notes: dbc.text
    meta: dbc.json
    disabled: bool = False


# Init Objects
dbc.load([UserMongo])

@pytest.fixture(scope="session")
def event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
    yield loop
    loop.close()


class ID:
    @classmethod
    async def one(cls):
        results = await UserMongo.objects.all()
        return results.data[0].id


@pytest.mark.asyncio
async def test_create():
    results = await UserMongo.objects.create({"name": "jane doll"})
    assert results.error == False and results.data.name == "jane doll"


@pytest.mark.asyncio
async def test_update():
    selector = await ID.one()
    form = {
        "name": "joe doe",
    }
    results = await UserMongo.objects.update(selector, form)
    assert results.error == False and results.data.name == "joe doe"


@pytest.mark.asyncio
async def test_detail():
    selector = await ID.one()
    results = await UserMongo.objects.detail(selector)
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_get_by():
    results = await UserMongo.objects.get_by(name="joe doe")
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_find_one():
    query = {"name": {"$regex": "joe"}}
    results = await UserMongo.objects.find_one(query)
    assert results.name == "joe doe"


@pytest.mark.asyncio
async def test_delete():
    selector = await ID.one()
    results = await UserMongo.objects.delete(selector)
    assert results.error == False and results.count == 1


@pytest.mark.asyncio
async def test_all():
    await UserMongo.objects.create([{"name": "joe doe"}, {"name": "jane doll"}])
    results = await UserMongo.objects.all()
    assert (
        results.error == False
        and isinstance(results.data, list) == True
        and len(results.data) == 2
    )


@pytest.mark.asyncio
async def test_filter_by():
    query = {"name": "joe doe"}
    results = await UserMongo.objects.filter_by(search=query, page=1, limit=100, sort_by="-id")
    assert results.error == False and results.count == 1


@pytest.mark.asyncio
async def test_find():
    search = [{"name": {"$regex": "joe"}}, {"name": {"$regex": "jane"}}]
    query = {"$or": search}
    results = await UserMongo.objects.find(query, page=1, limit=100, sort_by="-id")
    assert results.error == False and results.count == 2


@pytest.mark.asyncio
async def test_search():
    search_value = "j"
    columns = ["name", "notes"]
    results = await UserMongo.objects.search(
        columns=columns, value=search_value, page=1, limit=100, sort_by="-id"
    )
    assert results.error == False and results.count == 2


@pytest.mark.asyncio
async def test_reset():
    results = await UserMongo.objects.delete(None, all=True)
    assert results.error == False


@pytest.mark.asyncio
async def demo_example():
    selector = await ID.one()
    form = {
        "name": "joe doe",
    }
    # results = await UserMongo.objects.update(selector, form)
    results = await UserMongo.objects.update(selector, form)
    print(results)


# asyncio.run(demo_example())



"""
    Lib Utils
"""


def to_camel_case(text):
    """Converts to { camel-Case }"""
    init, *temp = text.split("_")
    return "".join([init.lower(), *map(str.title, temp)])


def to_pascal_case(text):
    """Converts to { Pascal-Case }"""
    temp = text.split("_")
    return "".join([*map(str.title, temp)])





from typing import Any


def clean_form(base: Any, cols: list, form: dict):
    """Clean User's Input

    Args:
        form (dict): User's Input.

    Returns:
        dict: Clean User's Input.
    """
    inputs = {}
    base_form = base(**form)
    for key, val in base_form.__dict__.items():
        if key in cols and key != "_id":
            inputs[key] = val
    return inputs


def clean_update_form(base: Any, cols: list, form: dict):
    """Clean User's Input (Update-Form)"""
    clean = clean_form(base, cols, form)
    return {key: val for key, val in clean.items() if val and key in form.keys()}



"""
    Databases Manager (SQL & MongoDB)
"""
import functools
from collections import namedtuple

from .. import manager
from ..manager.sql import Database as SQLDatabase
from .model import Model

try:
    import motor.motor_asyncio

    MotorClient = motor.motor_asyncio.AsyncIOMotorClient
except ImportError:
    MotorClient = False

try:
    from sqlalchemy.orm import declarative_base

except ImportError:
    declarative_base = False


DBManager = namedtuple("DBManager", ["base", "client", "database"])
DBController = namedtuple("DBController", ["sql", "mongo", "keys"])
Controller = namedtuple("Controller", ["sql", "mongo"])
MongoDNS = namedtuple("MongoDNS", ["url", "database"])


def get_database_name(url):
    """Get Table's Name from URL"""
    items = url.split("/")[-1::]
    return items[0] if len(items) == 1 else None


def get_database_url(url):
    """Get Database's URL"""
    return "/".join(url.split("/")[:-1])


def get_database_info(url):
    """{ get_database_name } + { get_database_url }"""
    return MongoDNS(url=get_database_url(url), database=get_database_name(url))


def databases_setup(sql: str = None, mongo: str = None):
    """Databases Setup"""
    # SQL
    sql_manager = DBManager(base=None, client=None, database=None)
    if declarative_base and sql:
        sql_base = declarative_base()
        sql_client = functools.partial(manager.SQL, sql)
        sql_manager = DBManager(
            base=sql_base, client=sql_client, database=SQLDatabase(sql)
        )

    # Mongo
    mongo_manager = DBManager(base=None, client=None, database=None)
    if MotorClient and mongo:
        mongo_dns = get_database_info(mongo)
        mongo_core = MotorClient(mongo_dns.url)
        # To Be Used
        mongo_base = mongo_core[mongo_dns.database]
        mongo_client = manager.Mongo
        mongo_manager = DBManager(
            base=mongo_base, client=mongo_client, database=mongo_base
        )

    return Controller(sql=sql_manager, mongo=mongo_manager)


class Database:
    """Database Models"""

    def __init__(self, sql: str = None, mongo: str = None, fastberry: bool = False):
        # Init
        self.fastberry = fastberry
        self._core_models = {}
        self._managers = {}
        self._model = Model()
        self._manager_sql = None
        self._manager_mongo = None
        self._base = None
        # Config Databases
        self._config(sql=sql, mongo=mongo)

    def _config(self, sql: str = None, mongo: str = None):
        """Config Databases"""
        # Core
        type_base = Model()
        app_managers = {
            "sql": type_base.type,
            "mongo": type_base.type,
        }
        the_manager = databases_setup(sql=sql, mongo=mongo)
        self._manager_sql = the_manager.sql
        self._manager_mongo = the_manager.mongo
        self._base = the_manager.sql.base
        if declarative_base:
            active = Model(sql=the_manager.sql.base)
            app_managers["sql"] = active.sql
        if MotorClient:
            active = Model(mongo=the_manager.mongo.base)
            app_managers["mongo"] = active.mongo
        # self._model = Model(sql=the_manager.sql.base, mongo=the_manager.mongo.base)
        self._model = Controller(**app_managers)

    def register(self, all_models: list):
        """Register a Type(Model)"""
        if not isinstance(all_models, list):
            all_models = [all_models]
        for current_type in all_models:
            self._core_models[current_type.__meta__.table_uri] = current_type

    def load(self):
        """Load Lazy-Tables"""
        for current_type in self._core_models.values():
            if current_type._lazy_object:
                if callable(current_type.objects):
                    current_type.objects()

    def set_fastberry(self, mode: bool = True):
        """Set Fastberry"""
        self.fastberry = mode

    @property
    def base(self):
        """SQLAlchemy Base"""
        return self._base

    @property
    def db(self):
        """Databases SQL & Mongo"""
        return self.database

    @property
    def database(self):
        """Databases SQL & Mongo"""
        return Controller(
            sql=self._manager_sql.database,
            mongo=self._manager_mongo.database,
        )

    @property
    def types(self):
        """Types"""
        return self._core_models

    @property
    def model(self):
        """Base SQL & Mongo"""
        return self._model

    @property
    def manager_sql(self):
        """Manager SQL"""
        return self._manager_sql

    @property
    def manager_mongo(self):
        """Manager Mongo"""
        return self._manager_mongo

    @property
    def manager(self):
        """Manager Mongo"""
        return self._managers

    def manage(self, all_models: list):
        """Register + Load"""
        if not self.fastberry:
            self.register(all_models)
            self.load()
        # Dicts
        manager_sql = {}
        manager_mongo = {}
        for current_type in all_models:
            if hasattr(current_type, "__meta__"):
                name = current_type.__meta__.table_uri
                is_sql = current_type.__meta__.sql
                is_mongo = current_type.__meta__.mongo
                if is_sql and declarative_base:
                    manager_sql[name] = self.manager_sql.client(current_type)
                elif is_mongo and MotorClient:
                    manager_mongo[name] = self.manager_mongo.client(current_type)

        all_keys = list(manager_sql.keys()) + list(manager_mongo.keys())
        db_manager = DBController(
            sql=manager_sql, mongo=manager_mongo, keys=lambda: all_keys
        )
        self._managers = db_manager
        return db_manager



"""
    Database Controller
"""
import dataclasses as dc
import functools
import typing

from .dbmanager import Database


@dc.dataclass(frozen=True)
class DatabaseController:
    """Single Controller"""

    name: str
    engine: str
    model: typing.Any
    admin: typing.Any
    database: typing.Any
    base: typing.Any = None


@dc.dataclass(frozen=True)
class Controller:
    """Full Controller"""

    sql: typing.Any
    mongo: typing.Any


def create_database_manager(
    all_models: list, engine_type: str = None, active_db: typing.Any = None
):
    """Create a { Database Manager }"""
    if not isinstance(all_models, list):
        all_models = [all_models]
    match engine_type:
        case "sql":
            manager = active_db.manage(all_models).sql  # .database .base
        case "mongo":
            manager = active_db.manage(all_models).mongo
    return manager


def create_database(name: str, engine_type: str, url: str, fastberry: bool = False):
    """Create a { Database }"""
    active_database = None
    custom_database = None
    match engine_type:
        case "sql":
            active_database = Database(sql=url, fastberry=fastberry)
            active_admin = functools.partial(
                create_database_manager,
                engine_type=engine_type,
                active_db=active_database,
            )
            custom_database = DatabaseController(
                name=name,
                engine=engine_type,
                admin=active_admin,
                model=active_database.model.sql,
                database=active_database.database.sql,
                base=active_database.base,
            )
        case "mongo":
            active_database = Database(mongo=url, fastberry=fastberry)
            active_admin = functools.partial(
                create_database_manager,
                engine_type=engine_type,
                active_db=active_database,
            )
            custom_database = DatabaseController(
                name=name,
                engine=engine_type,
                admin=active_admin,
                model=active_database.model.mongo,
                database=active_database.database.mongo,
            )
    return custom_database


def process_databases(
    databases: dict, engine_type: str = "sql", fastberry: bool = False
):
    """Process all { Databases }"""
    active_db = databases.get(engine_type)
    all_dbs = {}
    if active_db:
        for db_name, db_url in active_db.items():
            manager = create_database(db_name, engine_type, db_url, fastberry)
            all_dbs[db_name] = manager
    return all_dbs


def create_controllers(databases, fastberry: bool = False):
    """Controller Builder"""
    mongo_dbs = process_databases(databases, "mongo", fastberry=fastberry)
    sql_dbs = process_databases(databases, "sql", fastberry=fastberry)
    return Controller(sql=sql_dbs, mongo=mongo_dbs)



"""
    Controller for <Mongo & SQL>
    Optional: [Strawberry-GraphQL]
"""

from . import tools
from .manager import SQL, Mongo
from .manager.utils import Decode, Objects
from .types import Admin, Database, Form, Model
from .types import create_controllers as Controller

# Data-Class: < Maker >
input = Form.input
search = Form.search
crud = Form.crud
form = Form.form

# Data-Class: < Fields >
field = Form.field
filters = Form.filters

# Scalars
ID = Model.id

# Scalar: JSON
JSON = Model.json
json = Model.json

# Scalar: Text(str)
Text = Model.text
text = Model.text



"""Test - Library
"""

import asyncio
import functools
import os
import pathlib
import sys

import pytest


def dir_up(depth):
    """Easy level-up folder(s)."""
    return sys.path.append(os.path.join(pathlib.Path(__file__).parents[depth], "src"))


# Append to (sys.path)
dir_up(1)


import functools

# Test
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base

import dbcontroller as dbc

# URL
DATABASE_URL = "sqlite:///data/example.db"

# Base
Base = declarative_base()

# Setup
engine = create_engine(DATABASE_URL, echo=True)

model = dbc.Model(sql=Base)
SQL = functools.partial(dbc.SQL, DATABASE_URL)

# Types
@model.sql(table_name="main_user")
class User:
    name: str
    notes: dbc.Text
    meta: dbc.JSON
    disabled: bool = False


table = SQL(User)

# Create-Tables
# Base.metadata.create_all(engine)

# Register - Models
dbc.Admin.register([User])  # ... <All Other Models>
dbc.Admin.load()

print(dbc.Admin.types)
